# Copyright (c) 2022 Zhaiyu Chen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
graph.py
----------

Adjacency graph of the cell complex.

Each cell in the complex is represented as a node in the graph.
In addition, two nodes (S and T) are appended in the graph,
representing foreground and background, respectively.
There are two kinds of edges in the graph: n-links and st-links.
An n-link exists in between of two adjacent cells.
An st-link connects every cell to S and to T.
"""

import time
from pathlib import Path

import networkx as nx
import numpy as np
from sage.all import RR
from scipy.spatial import ConvexHull

from .logger import attach_to_log

logger = attach_to_log()


class AdjacencyGraph:
    """
    Class Adjacency graph of the cell complex.
    """

    def __init__(self, graph=None, quiet=False):
        """
        Init AdjacencyGraph.

        Parameters
        ----------
        graph: None or networkx Graph
            Graph object
        quiet: bool
            Disable logging if set True
        """
        if quiet:
            logger.disabled = True

        self.graph = graph
        self.uid = list(graph.nodes) if graph else None  # passed graph.nodes are sorted
        self.reachable = None  # for outer surface extraction
        self.non_reachable = None
        self._cached_interfaces = {}

    def load_graph(self, filepath):
        """
        Load graph from an external file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to networkx graph
        """
        filepath = Path(filepath)
        if filepath.suffix == '.adjlist':
            logger.info('loading graph from {}'.format(filepath))
            self.graph = nx.read_adjlist(filepath)
            self.uid = self._sort_uid()  # loaded graph.nodes are unordered string
        else:
            raise NotImplementedError('file format not supported: {}'.format(filepath.suffix))

    def assign_weights_to_n_links(self, cells, attribute='area_overlap', normalise=True, factor=1.0, engine='Qhull',
                                  cache_interfaces=False):
        """
        Assign weights to edges between every cell.

        Parameters
        ----------
        cells: list of Polyhedra objects
            Polyhedra cells
        attribute: str
            Attribute to use for encoding n-links, options are 'radius_overlap',
            'area_overlap', 'vertices_overlap', 'area_misalign' and 'volume_difference'
        normalise: bool
            Normalise the attribute if set True
        factor: float
            Factor to multiply to the attribute
        engine: str
            Engine to compute convex hull
            'Qhull' is supported at the moment
        cache_interfaces: bool
            Cache interfaces if set True
        """

        radius = [None] * len(self.graph.edges)
        area = [None] * len(self.graph.edges)
        volume = [None] * len(self.graph.edges)
        num_vertices = [None] * len(self.graph.edges)

        if attribute == 'radius_overlap':
            for i, (m, n) in enumerate(self.graph.edges):
                # compute interface
                interface = cells[self._uid_to_index(m)].intersection(cells[self._uid_to_index(n)])
                if cache_interfaces:
                    self._cached_interfaces[m, n] = interface  # uid pair as key
                radius[i] = RR(interface.radius())

            for i, (m, n) in enumerate(self.graph.edges):
                max_radius = max(radius)
                # small (sum of) overlapping radius -> large capacity -> small cost -> cut here
                self.graph[m][n].update({'capacity': ((max_radius - radius[
                    i]) / max_radius if normalise else max_radius - radius[i]) * factor})

        elif attribute == 'area_overlap':
            for i, (m, n) in enumerate(self.graph.edges):
                # compute interface
                interface = cells[self._uid_to_index(m)].intersection(cells[self._uid_to_index(n)])
                if cache_interfaces:
                    self._cached_interfaces[m, n] = interface  # uid pair as key
                if engine == 'Qhull':
                    # 'volume' is the area of the convex hull when input points are 2-dimensional
                    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.ConvexHull.html
                    try:
                        area[i] = ConvexHull(interface.affine_hull_projection().vertices_list()).volume
                    except:
                        # degenerate floating-point
                        area[i] = RR(interface.affine_hull_projection().volume())
                else:
                    # slower computation
                    area[i] = RR(interface.affine_hull_projection().volume())

            for i, (m, n) in enumerate(self.graph.edges):
                max_area = max(area)
                # balloon term
                # small (sum of) overlapping area -> large capacity -> small cost -> cut here
                self.graph[m][n].update(
                    {'capacity': ((max_area - area[i]) / max_area if normalise else max_area - area[i]) * factor})

        elif attribute == 'vertices_overlap':
            # number of vertices on overlapping areas
            for i, (m, n) in enumerate(self.graph.edges):
                # compute interface
                interface = cells[self._uid_to_index(m)].intersection(cells[self._uid_to_index(n)])
                if cache_interfaces:
                    self._cached_interfaces[m, n] = interface  # uid pair as key
                num_vertices[i] = interface.n_vertices()

            for i, (m, n) in enumerate(self.graph.edges):
                max_vertices = max(num_vertices)
                # few number of vertices -> large capacity -> small cost -> cut here
                self.graph[m][n].update({'capacity': ((max_vertices - num_vertices[
                    i]) / max_vertices if normalise else max_vertices - num_vertices[i]) * factor})

        elif attribute == 'area_misalign':
            # area_misalign makes little sense as observed from the results
            logger.warning('attribute "area_misalign" is deprecated')

            # area of the mis-aligned regions from both cells
            for i, (m, n) in enumerate(self.graph.edges):
                # compute interface
                interface = cells[self._uid_to_index(m)].intersection(cells[self._uid_to_index(n)])
                if cache_interfaces:
                    self._cached_interfaces[m, n] = interface  # uid pair as key

                for facet_m in cells[self._uid_to_index(m)].facets():
                    for facet_n in cells[self._uid_to_index(n)].facets():
                        if facet_m.ambient_Hrepresentation()[0].A() == -facet_n.ambient_Hrepresentation()[0].A() and \
                                facet_m.ambient_Hrepresentation()[0].b() == -facet_n.ambient_Hrepresentation()[0].b():
                            # two facets coplanar
                            # area of the misalignment
                            if engine == 'Qhull':
                                area[i] = ConvexHull(
                                    facet_m.as_polyhedron().affine_hull_projection().vertices_list()).volume + ConvexHull(
                                    facet_n.as_polyhedron().affine_hull_projection().vertices_list()).volume - 2 * ConvexHull(
                                    interface.affine_hull_projection().vertices_list()).volume
                            else:
                                area[i] = RR(
                                    facet_m.as_polyhedron().affine_hull_projection().volume() +
                                    facet_n.as_polyhedron().affine_hull_projection().volume() -
                                    2 * interface.affine_hull_projection().volume())

            for i, (m, n) in enumerate(self.graph.edges):
                max_area = max(area)
                self.graph[m][n].update(
                    {'capacity': (area[i] / max_area if normalise else area[i]) * factor})

        elif attribute == 'volume_difference':
            # encourage partition between relatively a big cell and a small cell
            for i, (m, n) in enumerate(self.graph.edges):
                if engine == 'Qhull':
                    volume[i] = abs(ConvexHull(cells[self._uid_to_index(m)].vertices_list()).volume - ConvexHull(
                        cells[self._uid_to_index(n)].vertices_list()).volume) / max(
                        ConvexHull(cells[self._uid_to_index(m)].vertices_list()).volume,
                        ConvexHull(cells[self._uid_to_index(n)].vertices_list()).volume)
                else:
                    volume[i] = RR(
                        abs(cells[self._uid_to_index(m)].volume() - cells[self._uid_to_index(n)].volume()) / max(
                            cells[self._uid_to_index(m)].volume(), cells[self._uid_to_index(n)].volume()))

            for i, (m, n) in enumerate(self.graph.edges):
                max_volume = max(volume)
                # large difference -> large capacity -> small cost -> cut here
                self.graph[m][n].update(
                    {'capacity': (volume[i] / max_volume if normalise else volume[i]) * factor})

    def assign_weights_to_st_links(self, weights):
        """
        Assign weights to edges between each cell and the s-t nodes.

        Parameters
        ----------
        weights: dict
            Weights in respect to each node, can be the occupancy probability or the signed distance of the cells.
        """
        for i in self.uid:
            self.graph.add_edge(i, 's', capacity=weights[i])
            self.graph.add_edge(i, 't', capacity=1 - weights[i])  # make sure

    def cut(self):
        """
        Perform cutting operation.

        Returns
        ----------
        cut_value: float
            Cost of the cut
        reachable: list of int
            Reachable nodes from the S node
        """
        tik = time.time()
        cut_value, partition = nx.algorithms.flow.minimum_cut(self.graph, 's', 't')
        reachable, non_reachable = partition
        reachable.remove('s')
        non_reachable.remove('t')
        self.reachable = reachable
        self.non_reachable = non_reachable

        logger.info('cut performed: {:.2f} s'.format(time.time() - tik))
        logger.info('cut_value: {:.2f}'.format(cut_value))
        logger.info('number of extracted cells: {}'.format(len(reachable)))
        return cut_value, reachable

    @staticmethod
    def _sorted_vertex_indices(adjacency_matrix):
        """
        Return sorted vertex indices.

        Parameters
        ----------
        adjacency_matrix: matrix
            Adjacency matrix

        Returns
        -------
        sorted_: list of int
            Sorted vertex indices
        """
        pointer = 0
        sorted_ = [pointer]
        for _ in range(len(adjacency_matrix[0]) - 1):
            connected = np.where(adjacency_matrix[pointer])[0]  # two elements
            if connected[0] not in sorted_:
                pointer = connected[0]
                sorted_.append(connected[0])
            else:
                pointer = connected[1]
                sorted_.append(connected[1])
        return sorted_

    @staticmethod
    def _reorient_facets(facets):
        """
        Given a list of facets (each facet is a list of vertex indices),
        reorient them so that every pair of adjacent facets shares a common edge
        with opposite direction.

        Assumes the mesh is connected and facet 0 is correctly oriented.
        """
        # Build a mapping from an unordered edge (sorted tuple) to a list of (face_index, edge)
        edge_to_faces = {}
        for fi, face in enumerate(facets):
            n = len(face)
            for j in range(n):
                edge = (face[j], face[(j + 1) % n])
                key = tuple(sorted(edge))
                edge_to_faces.setdefault(key, []).append((fi, edge))

        oriented = [False] * len(facets)
        oriented[0] = True  # assume facet 0 is correctly oriented
        queue = [0]

        while queue:
            current = queue.pop(0)
            n = len(facets[current])
            for j in range(n):
                # Get the current facet's edge in its current order
                edge = (facets[current][j], facets[current][(j + 1) % n])
                key = tuple(sorted(edge))
                # For every facet sharing this edge:
                for neighbor_index, neighbor_edge in edge_to_faces.get(key, []):
                    if not oriented[neighbor_index]:
                        # For consistent orientation, the shared edge in neighbor should be the reverse of edge
                        if neighbor_edge == edge:
                            # If not, flip the neighbor facet
                            facets[neighbor_index] = facets[neighbor_index][::-1]
                        oriented[neighbor_index] = True
                        queue.append(neighbor_index)
        return facets

    def save_surface_obj(self, filepath, cells=None, engine='mesh'):
        """
        Save the outer surface to an OBJ file, from interfaces between cells being cut.

        Parameters
        ----------
        filepath : str or Path
            Filepath to save the OBJ file.
        cells : None or list of Polyhedra objects
            Polyhedra cells.
        engine : str
            Engine to extract surface. Allowed values are 'mesh' or 'soup' (polygon soup).
        """
        if not self.reachable:
            logger.error('no reachable cells. aborting')
            return
        if not self.non_reachable:
            logger.error('no unreachable cells. aborting')
            return
        if not self._cached_interfaces and not cells:
            logger.error('neither cached interfaces nor cells are available. aborting')
            return
        if engine not in {'mesh', 'soup'}:
            logger.error('engine can be "mesh" or "soup"')
            return

        # Initialize variables
        interfaces = []
        surface = None if engine == 'soup' else None  # surface is only used in 'fast' engine
        tik = time.time()

        # Loop over edges and process only those between a reachable and a non-reachable cell.
        for edge in self.graph.edges:
            if edge[0] in self.reachable and edge[1] in self.non_reachable:
                r, nr = edge[0], edge[1]
            elif edge[1] in self.reachable and edge[0] in self.non_reachable:
                r, nr = edge[1], edge[0]
            else:
                continue

            # Retrieve interface from cache if available; otherwise, compute intersection.
            if self._cached_interfaces:
                interface = self._cached_interfaces.get((r, nr)) or self._cached_interfaces.get((nr, r))
            else:
                interface = cells[self._uid_to_index(r)].intersection(cells[self._uid_to_index(nr)])
            interfaces.append(interface)

            # For the 'fast' engine, accumulate rendered solid surfaces.
            if engine == 'soup':
                rendered = interface.render_solid()
                surface = rendered if surface is None else surface + rendered

        # Build the OBJ file string depending on the chosen engine.
        if engine == 'soup':
            surface_obj = surface.obj_repr(surface.default_render_params())
            lines = []
            for obj in surface_obj:
                lines.append(obj[0])
                lines.extend(obj[2])
                lines.extend(obj[3])
            surface_str = "\n".join(lines)
        elif engine == 'mesh':
            # Compute facets for each interface.
            facets = [
                np.array(interface.vertices())[self._sorted_vertex_indices(interface.adjacency_matrix())]
                for interface in interfaces
            ]
            # Extract unique vertices while preserving order.
            flat_vertices = [tuple(vertex) for facet in facets for vertex in facet]
            unique_vertices = list(dict.fromkeys(flat_vertices))
            vertex_to_index = {vertex: idx for idx, vertex in enumerate(unique_vertices)}

            # Map each facet's vertices to indices.
            facets_ = [[vertex_to_index[tuple(vertex)] for vertex in facet] for facet in facets]
            # Reorient facets so that each adjacent pair shares a common edge in opposite directions.
            facets_ = self._reorient_facets(facets_)

            # Build OBJ lines.
            obj_lines = [
                "v {} {} {}".format(float(v[0]), float(v[1]), float(v[2])) for v in unique_vertices
            ]
            for facet in facets_:
                face_line = "f " + " ".join(str(idx + 1) for idx in facet)
                obj_lines.append(face_line)
            surface_str = "\n".join(obj_lines)

        logger.info('surface extracted: {:.2f} s'.format(time.time() - tik))

        # Ensure directory exists and write the OBJ file.
        filepath = Path(filepath)
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, 'w') as f:
            f.write(surface_str)

    def draw(self):
        """
        Draw the graph with nodes represented by their UID.
        """
        import matplotlib.pyplot as plt
        plt.subplot(121)
        nx.draw(self.graph, with_labels=True, font_weight='bold')
        plt.show()

    def _uid_to_index(self, query):
        """
        Convert index in the node list to that in the cell list.

        The rationale behind is #nodes == #cells (when a primitive is settled down).

        Parameters
        ----------
        query: int
            Query uid in the node list

        Returns
        -------
        as_int: int
            Query uid in the cell list
        """
        return self.uid.index(query)

    def _index_to_uid(self, query):
        """
        Convert index to node UID.

        Parameters
        ----------
        query: int
            Query index in the node list

        Returns
        -------
        as_int: int
            Node UID
        """
        return self.uid[query]

    def _sort_uid(self):
        """
        Sort UIDs for graph structure loaded from an external file.

        Returns
        -------
        as_list: list of int
            Sorted UIDs
        """
        return sorted([int(i) for i in self.graph.nodes])

    def to_indices(self, uids):
        """
        Convert UIDs to indices.

        Parameters
        ----------
        uids: list of int
            UIDs of nodes

        Returns
        -------
        as_list: list of int
            Indices of nodes
        """
        return [self._uid_to_index(i) for i in uids]
    
    def to_uids(self, indices):
        """
        Convert indices to UIDs.

        Parameters
        ----------
        indices: list of int
            Indices of nodes

        Returns
        -------
        as_list: list of int
            UIDs of nodes
        """
        return [self._index_to_uid(i) for i in indices]

    def to_dict(self, weights_list):
        """
        Convert a weight list to weight dict keyed by self.uid.

        Parameters
        ----------
        weights_list: list of
            Weight list

        Returns
        -------
        as_dict: dict
            Weight dict
        """
        return {self.uid[i]: weight for i, weight in enumerate(weights_list)}



from pathlib import Path

import numpy as np
import trimesh

from abspy import CellComplex

dir_tests = Path(__file__).parent


def example_cell_complex_from_planes():
    """
    Simple CellComplex construction from specified planes and bounds.
    """
    # start from two planes
    planes = np.array([[0, 1, 0, -50], [0, 0, 1, -50]])

    # specify the bounds
    bounds = np.array([[[0, 50, 0], [100, 50, 100]], [[0, 0, 50], [100, 50, 50]]])

    # specify the initial bound
    initial_bound = [[0, 0, 0], [100, 100, 100]]

    # initialise the cell complex
    cell_complex = CellComplex(planes, bounds, initial_bound=initial_bound, build_graph=True)

    # prioritise certain (vertical by default) planes
    cell_complex.prioritise_planes()

    # construct the complex
    cell_complex.construct()

    # print out info
    cell_complex.print_info()

    # boundary cells
    cells_boundary = cell_complex.cells_boundary()
    print(f'boundary cells: {cells_boundary}')

    # cell representatives
    num_representatives = 1000
    representatives = cell_complex.cell_representatives(location='skeleton', num=num_representatives)
    representatives = np.concatenate(representatives, axis=0)
    colours = np.zeros([num_representatives * cell_complex.num_cells, 4])
    for i in range(cell_complex.num_cells):
        colours[num_representatives * i: num_representatives * (i + 1)] = trimesh.visual.color.random_color()

    # visualise cell representatives
    representatives_vis = trimesh.PointCloud(representatives, colours)
    representatives_vis.export(dir_tests / 'test_output' / 'test_vis.ply', file_type='ply')
    try:
        representatives_vis.show()
    except (Exception, ImportError) as e:
        print(f'visualization skipped: {e}')

    # save cell complex CC file
    cell_complex.save(dir_tests / 'test_output' / 'complex.cc')

    # save cells to OBJ and PLM files
    cell_complex.save_obj(dir_tests / 'test_output' / 'cells.obj', use_mtl=True)
    cell_complex.save_plm(dir_tests / 'test_output' / 'cells.plm')


def example_cell_complex_from_mesh():
    """
    CellComplex construction from reference mesh.
    """
    # vertex group reference from mesh
    from abspy import VertexGroupReference
    vertex_group_reference = VertexGroupReference(filepath=dir_tests / 'test_data' / 'test_church.obj', num_samples=10000)

    # perturb plane normal vectors
    vertex_group_reference.perturb(sigma=0.001)

    # construct cell complex
    cell_complex = CellComplex(vertex_group_reference.planes, vertex_group_reference.aabbs,
                               vertex_group_reference.obbs, build_graph=True, quiet=False)
    cell_complex.construct()
    cell_complex.print_info()

    # cells inside reference mesh
    cells_in_mesh = cell_complex.cells_in_mesh(dir_tests / 'test_data' / 'test_church.obj', engine='distance')

    # save cell complex CC file
    cell_complex.save(dir_tests / 'test_output' / 'test_complex.cc')

    #  Save cells to OBJ and PLM files
    cell_complex.save_obj(dir_tests / 'test_output' / 'test_cells.obj', use_mtl=True, indices_cells=cells_in_mesh)
    cell_complex.save_plm(dir_tests / 'test_output' / 'test_cells.plm')

    # visualise the inside cells (only if pyglet installation is found and valid indices are provided)
    if len(cells_in_mesh):
        try:
            cell_complex.visualise(indices_cells=cells_in_mesh, temp_dir='./test_output/')
        except (Exception, ImportError) as e:
            print(f'visualization skipped: {e}')


if __name__ == '__main__':
    example_cell_complex_from_planes()
    example_cell_complex_from_mesh()




from pathlib import Path

from abspy import AdjacencyGraph


dir_tests = Path(__file__).parent


def example_graph():
    """
    Simple adjacency graph from external (.adjlist) file.
    """
    # initialise adjacency graph
    adjacency_graph = AdjacencyGraph()

    # load graph from file
    adjacency_graph.load_graph(dir_tests / 'test_data' / 'test_graph.adjlist')

    # simply draw the graph in 2D
    adjacency_graph.draw()


if __name__ == '__main__':
    example_graph()



# Copyright (c) 2022 Zhaiyu Chen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
primitive.py
----------

Process detected planar primitives.

Primitives are supported in vertex group format (.vg, .bvg).
Mapple as in [Easy3D](https://github.com/LiangliangNan/Easy3D)
can be used to generate such primitives from point clouds.
Otherwise, one can refer to the vertex group file format specification
attached to the README document.
"""

from random import random
from pathlib import Path
from functools import reduce
import struct

import numpy as np
from sklearn.neighbors import KDTree
from sklearn.decomposition import PCA

from .logger import attach_to_log

logger = attach_to_log()

try:
    import open3d as o3d
except ModuleNotFoundError:
    logger.warning('Open3D import failed')
    o3d = None


class VertexGroup:
    """
    Class for manipulating planar primitives.
    """

    def __init__(self, filepath, process=True, quiet=False, refit=True, global_group=False):
        """
        Init VertexGroup.
        Class for manipulating planar primitives.

        Parameters
        ----------
        filepath: str or Path
            Filepath to vertex group file (.vg) or binary vertex group file (.bvg)
        process: bool
            Immediate processing if set True
        quiet: bool
            Disable logging if set True
        refit: bool
            Refit plane parameters if set True
        global_group: bool
            Remove the first group as an unnecessary global one containing all subgroups if set True
        """
        if quiet:
            logger.disabled = True

        if isinstance(filepath, str):
            self.filepath = Path(filepath)
        else:
            self.filepath = filepath
        self.processed = False
        self.points = None
        self.planes = None
        self.aabbs = None
        self.obbs = None
        self.points_grouped = None
        self.points_ungrouped = None

        self.vgroup_ascii = self.load_file()
        self.vgroup_binary = None

        self.refit = refit
        self.global_group = global_group
        self.normals = None

        if process:
            self.process()

    def load_file(self):
        """
        Load (ascii / binary) vertex group file.
        """
        if self.filepath.suffix == '.vg':
            with open(self.filepath, 'r') as fin:
                return fin.readlines()

        elif self.filepath.suffix == '.bvg':
            # define size constants
            _SIZE_OF_INT = 4
            _SIZE_OF_FLOAT = 4
            _SIZE_OF_PARAM = 4
            _SIZE_OF_COLOR = 3

            vgroup_ascii = ''
            with open(self.filepath, 'rb') as fin:
                # points
                num_points = struct.unpack('i', fin.read(_SIZE_OF_INT))[0]
                points = struct.unpack('f' * num_points * 3, fin.read(_SIZE_OF_FLOAT * num_points * 3))
                vgroup_ascii += f'num_points: {num_points}\n'
                vgroup_ascii += ' '.join(map(str, points)) + '\n'

                # colors
                num_colors = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                vgroup_ascii += f'num_colors: {num_colors}\n'

                # normals
                num_normals = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                normals = struct.unpack('f' * num_normals * 3, fin.read(_SIZE_OF_FLOAT * num_normals * 3))
                vgroup_ascii += f'num_normals: {num_normals}\n'
                vgroup_ascii += ' '.join(map(str, normals)) + '\n'

                # groups
                num_groups = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                vgroup_ascii += f'num_groups: {num_groups}\n'

                group_counter = 0
                while group_counter < num_groups:
                    group_type = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                    num_group_parameters = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                    group_parameters = struct.unpack("f" * _SIZE_OF_PARAM, fin.read(_SIZE_OF_INT * _SIZE_OF_PARAM))
                    group_label_size = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                    # be reminded that vg <-> bvg in Mapple does not maintain group order
                    group_label = struct.unpack("c" * group_label_size, fin.read(group_label_size))
                    group_color = struct.unpack("f" * _SIZE_OF_COLOR, fin.read(_SIZE_OF_FLOAT * _SIZE_OF_COLOR))
                    group_num_point = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]
                    group_points = struct.unpack("i" * group_num_point, fin.read(_SIZE_OF_INT * group_num_point))
                    num_children = struct.unpack("i", fin.read(_SIZE_OF_INT))[0]

                    vgroup_ascii += f'group_type: {group_type}\n'
                    vgroup_ascii += f'num_group_parameters: {num_group_parameters}\n'
                    vgroup_ascii += 'group_parameters: ' + ' '.join(map(str, group_parameters)) + '\n'
                    vgroup_ascii += 'group_label: ' + ''.join(map(str, group_label)) + '\n'
                    vgroup_ascii += 'group_color: ' + ' '.join(map(str, group_color)) + '\n'
                    vgroup_ascii += f'group_num_point: {group_num_point}\n'
                    vgroup_ascii += ' '.join(map(str, group_points)) + '\n'
                    vgroup_ascii += f'num_children: {num_children}\n'

                    group_counter += 1

                # convert vgroup_ascii to list
                return vgroup_ascii.split('\n')

        else:
            raise ValueError(f'unable to load {self.filepath}, expected *.vg or .bvg.')

    def process(self):
        """
        Start processing vertex group.
        """
        logger.info('processing {}'.format(self.filepath))
        self.points = self.get_points()
        self.planes, self.aabbs, self.obbs, self.points_grouped, self.points_ungrouped = self.get_primitives()
        self.processed = True

    def get_points(self, row=1):
        """
        Get points from vertex group.

        Parameters
        ----------
        row: int
            Row number where points are specified, defaults to 1 for filename.vg

        Returns
        ----------
        as_float: (n, 3) float
            Point cloud
        """
        pc_lines = []

        for line in self.vgroup_ascii[row:]:
            # stop reading when the 'num_colors' keyword is found
            if 'num_colors' in line:
                break
            pc_lines.append(line)

        pc = np.fromstring(' '.join(pc_lines), sep=' ')
        return np.reshape(pc, (-1, 3))

    def get_primitives(self):
        """
        Get primitives from vertex group.

        Returns
        ----------
        params: (n, 4) float
            Plane parameters
        aabbs: (n, 2, 3) float
            Axis-aligned bounding boxes of the primitives
        obbs: (n, 4, 3) float
            Oriented bounding boxes of the primitives
        groups: (n, m, 3) float
            Groups of points
        ungrouped_points: (u, 3) float
            Points that belong to no group
        """
        is_primitive = [line.startswith('group_num_point') for line in self.vgroup_ascii]
        is_parameter = [line.startswith('group_parameters') for line in self.vgroup_ascii]

        primitives = [self.vgroup_ascii[line] for line in np.where(is_primitive)[0] + 1]  # lines of groups in the file
        parameters = [self.vgroup_ascii[line] for line in np.where(is_parameter)[0]]

        # remove global group if there is one
        if self.global_group:
            primitives = primitives[1:]
            parameters = parameters[1:]

        if len(primitives) != len(parameters):
            raise ValueError('group attributes mismatch')

        params = []
        aabbs = []
        obbs = []
        groups = []
        grouped_indices = set()  # indices of points being grouped
        for i, p in enumerate(primitives):
            point_indices = np.fromstring(p, sep=' ', dtype=np.int64)
            grouped_indices.update(point_indices)
            points = self.points[point_indices]

            if len(point_indices) == 0:
                # empty group -> global bounds and no refit
                if self.refit:
                    logger.warning('refit skipped for empty group')
                param = np.array([float(j) for j in parameters[i][18:-1].split()])
                aabb = self._points_bound(self.points)
                obb = aabb

            else:
                # valid group -> local bounds and optional refit
                if self.refit:
                    param, obb = self.fit_plane(points, mode='PCA')
                else:
                    param = np.array([float(j) for j in parameters[i][18:-1].split()])
                    _, obb = self.fit_plane(points, mode='PCA')
                aabb = self._points_bound(points)

            if param is None or len(param) != 4:
                logger.warning(f'bad parameter skipped: {param}')
                continue

            params.append(param)
            aabbs.append(aabb)
            obbs.append(obb)
            groups.append(points)

        ungrouped_indices = set(range(len(self.points))).difference(grouped_indices)
        ungrouped_points = self.points[list(ungrouped_indices)]  # points that belong to no groups
        return (np.array(params), np.array(aabbs), np.array(obbs), np.array(groups, dtype=object),
                np.array(ungrouped_points))

    @staticmethod
    def _points_bound(points):
        """
        Get bounds (AABB) of the points.

        Parameters
        ----------
        points: (n, 3) float
            Points
        Returns
        ----------
        as_float: (2, 3) float
            Bounds (AABB) of the points
        """
        return np.array([np.amin(points, axis=0), np.amax(points, axis=0)])

    def normalise_from_centroid_and_scale(self, centroid, scale, num=None):
        """
        Normalising points.

        Centroid and scale are provided to be mitigated, which are identical with the return of
        scale_and_offset() such that the normalised points align with the corresponding mesh.
        Notice the difference with normalise_points_to_centroid_and_scale().

        Parameters
        ----------
        centroid: (3,) float
            Centroid of the points to be mitigated
        scale: float
            Scale of the points to be mitigated
        num: None or int
            If specified, random sampling is performed to ensure the identical number of points

        Returns
        ----------
        None: NoneType
            Normalised (and possibly sampled) self.points
        """
        # mesh_to_sdf.utils.scale_to_unit_sphere()
        self.points = (self.points - centroid) / scale

        # update planes and bounds as point coordinates has changed
        self.planes, self.aabbs, self.obbs, self.points_grouped, self.points_ungrouped = self.get_primitives()

        # safely sample points after planes are extracted
        if num:
            choice = np.random.choice(self.points.shape[0], num, replace=True)
            self.points = self.points[choice, :]

    def normalise_to_centroid_and_scale(self, centroid=(0, 0, 0), scale=1.0, num=None):
        """
        Normalising points to the provided centroid and scale. Notice
        the difference with normalise_points_from_centroid_and_scale().

        Parameters
        ----------
        centroid: (3,) float
            Desired centroid of the points
        scale: float
            Desired scale of the points
        num: None or int
            If specified, random sampling is performed to ensure the identical number of points

        Returns
        ----------
        None: NoneType
            Normalised (and possibly sampled) self.points
        """
        ######################################################
        # this does not lock the scale
        # offset = np.mean(points, axis=0)
        # denominator = np.max(np.ptp(points, axis=0)) / scale
        ######################################################
        bounds = np.ptp(self.points, axis=0)
        center = np.min(self.points, axis=0) + bounds / 2
        offset = center
        self.points = (self.points - offset) / (bounds.max() * scale) + centroid

        # update planes and bounds as point coordinates has changed
        self.planes, self.aabbs, self.obbs, self.points_grouped, self.points_ungrouped = self.get_primitives()

        # safely sample points after planes are extracted
        if num:
            choice = np.random.choice(self.points.shape[0], num, replace=True)
            self.points = self.points[choice, :]

    @staticmethod
    def fit_plane(points, mode='PCA'):
        """
        Fit plane parameters for a point set.

        Parameters
        ----------
        points: (n, 3) float
            Points to be fit
        mode: str
            Mode of plane fitting,
            'PCA' (recommended) or 'LSA' (may introduce distortions)

        Returns
        ----------
        param: (4,) float
            Plane parameters, (a, b, c, d) as in a * x + b * y + c * z = -d
        obb: (4,3) float
            Oriented bounding box of the plane
        """
        assert mode == 'PCA' or mode == 'LSA'

        if len(points) < 3:
            logger.warning('plane fitting skipped given #points={}'.format(len(points)))
            return None

        if mode == 'LSA':
            # AX = B
            logger.warning('LSA introduces distortions when the plane crosses the origin')
            param = np.linalg.lstsq(points, np.expand_dims(np.ones(len(points)), 1))
            param = np.append(param[0], -1)
            obb = None

        else:
            # PCA followed by shift
            pca = PCA(n_components=3)
            pca.fit(points)
            eig_vec = pca.components_
            points_trans = pca.transform(points)
            point_min = np.amin(points_trans, axis=0)
            point_max = np.amax(points_trans, axis=0)
            obb = np.array([[point_min[0], point_min[1], 0], [point_min[0], point_max[1], 0],
                            [point_max[0], point_max[1], 0], [point_max[0], point_min[1], 0]])
            obb = pca.inverse_transform(obb)

            logger.debug('explained_variance_ratio: {}'.format(pca.explained_variance_ratio_))

            # normal vector of minimum variance
            normal = eig_vec[2, :]  # (a, b, c) normalized
            centroid = np.mean(points, axis=0)

            # every point (x, y, z) on the plane satisfies a * x + b * y + c * z = -d
            # taking centroid as a point on the plane
            d = -centroid.dot(normal)
            param = np.append(normal, d)

        return param, obb

    def append_planes(self, additional_planes, additional_points=None):
        """
        Append planes to vertex group. The provided planes can be accompanied by optional supporting points.
        Notice these additional planes differ from `additional_planes` in `complex.py`: the former apply to
        the VertexGroup data structure thus is generic to applications, while the latter apply to only the
        CellComplex data structure.

        Parameters
        ----------
        additional_planes: (m, 4) float
            Plane parameters
        additional_points: None or (m, n, 3) float
            Points that support planes
        """
        if additional_points is None:
            # this may still find use cases where plane parameters are provided as-is
            logger.warning('no supporting points provided. only appending plane parameters')
        else:
            assert len(additional_planes) == len(additional_points)
            # direct appending would not work
            combined = np.zeros(len(self.points_grouped) + len(additional_points), dtype=object)
            combined[:len(self.points_grouped)] = self.points_grouped
            combined[len(self.points_grouped):] = [np.array(g) for g in additional_points]
            self.points_grouped = combined
        self.planes = np.append(self.planes, additional_planes, axis=0)

    def save_vg(self, filepath):
        """
        Save vertex group into a vg file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save vg file
        """
        logger.info('writing vertex group into {}'.format(filepath))

        if isinstance(filepath, str):
            assert filepath.endswith('.vg')
        elif isinstance(filepath, Path):
            assert filepath.suffix == '.vg'
        assert self.planes is not None and self.points_grouped is not None

        points_grouped = np.concatenate(self.points_grouped)
        points_ungrouped = self.points_ungrouped

        # points
        out = ''
        out += 'num_points: {}\n'.format(len(points_grouped) + len(points_ungrouped))
        for i in points_grouped.flatten():
            # https://stackoverflow.com/questions/54367816/numpy-np-fromstring-not-working-as-hoped
            out += '{} '.format(i)
        for i in points_ungrouped.flatten():
            out += '{} '.format(i)

        # colors (no color needed)
        out += '\nnum_colors: {}'.format(0)

        # normals
        out += '\nnum_normals: {}\n'.format(len(points_grouped) + len(points_ungrouped))
        for i, group in enumerate(self.points_grouped):
            for _ in group:
                out += '{} {} {} '.format(*self.planes[i][:3])
        for i in range(len(points_ungrouped)):
            out += '{} {} {} '.format(0, 0, 0)

        # groups
        num_groups = len(self.points_grouped)
        out += '\nnum_groups: {}\n'.format(num_groups)
        j_base = 0
        for i in range(num_groups):
            out += 'group_type: {}\n'.format(0)
            out += 'num_group_parameters: {}\n'.format(4)
            out += 'group_parameters: {} {} {} {}\n'.format(*self.planes[i])
            out += 'group_label: group_{}\n'.format(i)
            out += 'group_color: {} {} {}\n'.format(random(), random(), random())
            out += 'group_num_point: {}\n'.format(len(self.points_grouped[i]))
            for j in range(j_base, j_base + len(self.points_grouped[i])):
                out += '{} '.format(j)
            j_base += len(self.points_grouped[i])
            out += '\nnum_children: {}\n'.format(0)

        # additional groups without supporting points
        num_planes = len(self.planes)  # num_planes >= num_groups
        for i in range(num_groups, num_planes):
            out += 'group_type: {}\n'.format(0)
            out += 'num_group_parameters: {}\n'.format(4)
            out += 'group_parameters: {} {} {} {}\n'.format(*self.planes[i])
            out += 'group_label: group_{}\n'.format(i)
            out += 'group_color: {} {} {}\n'.format(random(), random(), random())
            out += 'group_num_point: {}\n'.format(0)
            out += 'num_children: {}\n'.format(0)

        with open(filepath, 'w') as fout:
            fout.writelines(out)

    def save_bvg(self, filepath):
        """
        Save vertex group into a bvg file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save vg file
        """
        logger.info('writing vertex group into {}'.format(filepath))

        if isinstance(filepath, str):
            assert filepath.endswith('.bvg')
        elif isinstance(filepath, Path):
            assert filepath.suffix == '.bvg'
        assert self.planes is not None and self.points_grouped is not None

        points_grouped = np.concatenate(self.points_grouped)
        points_ungrouped = self.points_ungrouped

        # points
        out = [struct.pack('i', len(points_grouped) + len(points_ungrouped))]
        for i in points_grouped.flatten():
            # https://stackoverflow.com/questions/54367816/numpy-np-fromstring-not-working-as-hoped
            out.append(struct.pack('f', i))
        for i in points_ungrouped.flatten():
            out.append(struct.pack('f', i))

        # colors (no color needed)
        out.append(struct.pack('i', 0))

        # normals
        out.append(struct.pack('i', len(points_grouped) + len(points_ungrouped)))
        for i, group in enumerate(self.points_grouped):
            for _ in group:
                out.append(struct.pack('fff', *self.planes[i][:3]))
        for i in range(len(points_ungrouped)):
            out.append(struct.pack('fff', 0, 0, 0))

        # groups
        num_groups = len(self.points_grouped)
        out.append(struct.pack('i', num_groups))
        j_base = 0
        for i in range(num_groups):
            out.append(struct.pack('i', 0))
            out.append(struct.pack('i', 4))
            out.append(struct.pack('ffff', *self.planes[i]))
            out.append(struct.pack('i', 6 + len(str(i))))
            out.append(struct.pack(f'{(6 + len(str(i)))}s', bytes('group_{}'.format(i), encoding='ascii')))
            out.append(struct.pack('fff', random(), random(), random()))
            out.append(struct.pack('i', len(self.points_grouped[i])))

            for j in range(j_base, j_base + len(self.points_grouped[i])):
                out.append(struct.pack('i', j))

            j_base += len(self.points_grouped[i])
            out.append(struct.pack('i', 0))

        # additional groups without supporting points
        num_planes = len(self.planes)  # num_planes >= num_groups
        for i in range(num_groups, num_planes):
            out.append(struct.pack('i', 0))
            out.append(struct.pack('i', 4))
            out.append(struct.pack('ffff', *self.planes[i]))
            out.append(struct.pack('i', 6 + len(str(i))))
            out.append(struct.pack(f'{(6 + len(str(i)))}s', bytes('group_{}'.format(i), encoding='ascii')))
            out.append(struct.pack('fff', random(), random(), random()))
            out.append(struct.pack('i', 0))
            out.append(struct.pack('i', 0))

        with open(filepath, 'wb') as fout:
            fout.writelines(out)

    def save_planes_txt(self, filepath):
        """
        Save plane parameters into a txt file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save txt file
        """
        with open(filepath, 'w') as fout:
            logger.info('writing plane parameters into {}'.format(filepath))
            out = [''.join(str(n) + ' ' for n in line.tolist()) + '\n' for line in self.planes]
            fout.writelines(out)

    def save_planes_npy(self, filepath):
        """
        Save plane params into an npy file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save npy file
        """
        logger.info('writing plane parameters into {}'.format(filepath))
        np.save(filepath, self.planes)

    def save_aabbs_npy(self, filepath):
        """
        Save plane AABBs into an npy file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save npy file
        """
        logger.info('writing plane bounds into {}'.format(filepath))
        np.save(filepath, self.aabbs)

    def save_cloud(self, filepath):
        """
        Save point cloud into a common 3D format. Support formats: xyzn, xyzrgb, pts, ply, pcd.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save point cloud file
        """
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(self.points)
        if self.normals is not None:
            pcd.normals = o3d.utility.Vector3dVector(self.normals)
        o3d.io.write_point_cloud(str(filepath), pcd)


class VertexGroupReference:
    """
    Class of reference vertex group sampled from meshes.
    """

    def __init__(self, filepath, num_samples=10000, process=True, quiet=False):
        """
        Init VertexGroupReference.
        Class of reference vertex group sampled from meshes.

        Parameters
        ----------
        filepath: str or Path
            Filepath to a mesh
        num_samples: int
            Number of sampled points
        process: bool
            Immediate processing if set True
        quiet: bool
            Disable logging if set True
        """
        if quiet:
            logger.disabled = True
        import trimesh

        self.filepath = filepath
        self.num_samples = num_samples
        self.processed = False
        self.points = None
        self.planes = None
        self.aabbs = None
        self.obbs = None
        self.points_grouped = None

        self.mesh = trimesh.load_mesh(self.filepath)
        self.normals = None

        if process:
            self.process()

    @staticmethod
    def _points_bound(points):
        """
        Get bounds (AABB) of the points.

        Parameters
        ----------
        points: (n, 3) float
            Points

        Returns
        ----------
        as_float: (2, 3) float
            Bounds (AABB) of the points
        """
        return np.array([np.amin(points, axis=0), np.amax(points, axis=0)])

    def process(self):
        """
        Start processing mesh data.
        """
        logger.info('processing {}'.format(self.filepath))

        # sample on all faces
        samples, face_indices = self.mesh.sample(count=self.num_samples, return_index=True)  # face_indices match facets

        planes = []
        aabbs = []
        obbs = []
        groups = []

        for facet in self.mesh.facets:  # a list of face indices for coplanar adjacent faces
            # group corresponding samples by facet
            points = []
            for face_index in facet:
                sample_indices = np.where(face_indices == face_index)[0]
                if len(sample_indices) > 0:
                    points.append(samples[sample_indices])

            # vertices
            vertices = reduce(np.union1d, self.mesh.faces[facet])  # indices of vertices
            vertices = self.mesh.vertices[vertices]  # coordinates of vertices

            # append vertices in case there is no sampled points in this group
            points.append(vertices)
            points = np.concatenate(points)

            # calculate parameters
            plane, obb = VertexGroup.fit_plane(vertices)

            planes.append(plane)
            aabbs.append(self._points_bound(vertices))
            obbs.append(obb)
            groups.append(points)

        # self.mesh.facets do not cover all faces
        faces_extracted = np.concatenate(self.mesh.facets)
        faces_remainder = np.setdiff1d(np.arange(len(self.mesh.faces)), faces_extracted)

        for face_index in faces_remainder:
            # group corresponding samples by faces
            points = []
            sample_indices = np.where(face_indices == face_index)[0]
            if len(sample_indices) > 0:
                points.append(samples[sample_indices])

            # vertices
            vertices = self.mesh.faces[face_index]
            vertices = self.mesh.vertices[vertices]

            # append vertices in case there is no sampled points in this group
            points.append(vertices)
            points = np.concatenate(points)

            # calculate parameters
            plane, obb = VertexGroup.fit_plane(vertices)
            planes.append(plane)
            aabbs.append(self._points_bound(vertices))
            obbs.append(obb)
            groups.append(points)

        self.points = np.concatenate(groups)
        self.planes = np.array(planes)
        self.aabbs = np.array(aabbs)
        self.obbs = np.array(obbs)
        self.points_grouped = np.array(groups, dtype=object)
        self.processed = True

    @property
    def bottom_indices(self, epsilon=0.01):
        """
        Group indices of bottom facets.

        Parameters
        ----------
        epsilon: (1,) float
            Tolerance for horizontality and minimum Z predicates

        Returns
        ----------
        as_int: (n,) int
            Indices of bottom groups
        """
        is_horizontal = np.logical_and(np.abs(self.planes[:, 0]) < epsilon, np.abs(self.planes[:, 1]) < epsilon)
        horizontal_indices = np.where(is_horizontal)[0]

        z = -self.planes[is_horizontal][:, 2] * self.planes[is_horizontal][:, 3]
        is_bottom = z < (np.min(z) + epsilon)

        return horizontal_indices[is_bottom]

    @property
    def wall_indices(self, epsilon=0.01):
        """
        Group indices of wall facets.

        Parameters
        ----------
        epsilon: (1,) float
            Tolerance for verticality predicate

        Returns
        ----------
        as_int: (n,) int
            Indices of wall groups
        """
        wall_indices = []
        is_vertical = np.abs(self.planes[:, 2]) < epsilon
        vertical_indices = np.where(is_vertical)[0]

        num_facets = len(self.mesh.facets)
        faces_remainder = np.setdiff1d(np.arange(len(self.mesh.faces)), np.concatenate(self.mesh.facets))

        # extract bottom vertices
        vertices_bottom = set()  # indices of bottom vertices
        for b in self.bottom_indices:
            if b < num_facets:
                faces = self.mesh.faces[self.mesh.facets[b]]
                vertices_bottom.update(reduce(np.union1d, faces))
            else:
                vertices_bottom.update(self.mesh.faces[faces_remainder[b - num_facets]])

        # extract vertical indices
        for v in vertical_indices:
            if v < num_facets:
                faces = self.mesh.faces[self.mesh.facets[v]]
                vertices_vertical = reduce(np.union1d, faces)  # indices of vertices
            else:
                vertices_vertical = self.mesh.faces[faces_remainder[v - num_facets]]  # indices of vertices

            if set(vertices_vertical).intersection(vertices_bottom):
                wall_indices.append(v)
        return np.array(wall_indices)

    def perturb(self, sigma):
        """
        Perturb plane normals with Gaussian noise.

        Parameters
        ----------
        sigma: (1,) float
            Gaussian noise standard deviation
        """
        # length of noise vector
        l = np.abs(np.random.normal(loc=0, scale=sigma, size=self.planes.shape[0]))

        # normal noise
        na = np.random.normal(loc=0, scale=1, size=self.planes.shape[0])
        nb = np.random.normal(loc=0, scale=1, size=self.planes.shape[0])
        nc = np.random.normal(loc=0, scale=1, size=self.planes.shape[0])

        # normalized noise vector
        n = np.array([na, nb, nc]).T / np.sqrt(na ** 2 + nb ** 2 + nc ** 2)[:, np.newaxis] * l[:, np.newaxis]

        # apply noise
        self.planes[:, :3] += n

        # re-normalize normal
        norms = np.linalg.norm(self.planes[:, :3], axis=1)
        self.planes[:, :3] /= norms[:, np.newaxis]

    def inject_points(self, points, threshold=0.05, overwrite=True, keep_bottom=False, keep_wall=False,
                      compute_normal=False, pseudo_normal=False, pseudo_size=1):
        """
        Inject unordered points to vertex groups.

        Parameters
        ----------
        points: (n, 3) float
            Point cloud
        threshold: (1,) float
            Distance threshold
        overwrite: bool
            Overwrite sampled points with input points if set True, otherwise append
        keep_bottom: bool
            Keep sampled points on bottom if set True, effective only when overwrite is True
        keep_wall: bool
            Keep sampled points on walls if set True, effective only when overwrite is True
        compute_normal: bool
            Compute normal if set True, otherwise inherit normal from plane parameters
        pseudo_normal: bool
            Take pseudo points into account for normal estimation if set True, otherwise random normal for them,
            not implemented
        pseudo_size: (1,) int
            Minimal group size for pseudo group without injected points,
            only apply to dangling groups without injected points, effective only when overwrite is True
        """
        assert self.points is not None

        # build KD-tree
        kdtree = KDTree(self.points)

        # query the KD-tree to find nearest neighbors and minimal distances
        min_distances, min_reference_indices = kdtree.query(points, k=1)
        min_distances = min_distances.flatten()
        min_reference_indices = min_reference_indices.flatten()

        # determine the group for each minimal reference point index
        group_boundaries = np.cumsum([0] + [chunk.shape[0] for chunk in self.points_grouped])
        min_reference_groups = np.digitize(min_reference_indices, group_boundaries) - 1
        min_reference_groups[min_distances > threshold] = -1

        # append points to groups (append, overwrite, keep all, or keep minimal)
        pseudo_groups = []
        for i in range(len(self.points_grouped)):
            group_mask = min_reference_groups == i
            if not overwrite:
                self.points_grouped[i] = np.concatenate([self.points_grouped[i], points[group_mask]], axis=0)
            elif (keep_bottom and i in self.bottom_indices) or (keep_wall and i in self.wall_indices):
                pass
            elif any(group_mask):
                self.points_grouped[i] = points[group_mask]
            else:
                self.points_grouped[i] = self.points_grouped[i][:pseudo_size]
                pseudo_groups.append(i)

        # update points and optionally their normals
        self.points = np.concatenate(self.points_grouped)
        self.points = np.concatenate([self.points, points[min_reference_groups == -1]])
        if compute_normal:
            pcd = o3d.geometry.PointCloud()
            pcd.points = o3d.utility.Vector3dVector(self.points)
            pcd.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(
                radius=4, max_nn=300))
            pcd.orient_normals_consistent_tangent_plane(k=15)
            self.normals = np.asarray(pcd.normals)

    def save_vg(self, filepath):
        """
        Save primitives into a vg file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save vg file
        """
        if isinstance(filepath, str):
            assert filepath.endswith('.vg')
        elif isinstance(filepath, Path):
            assert filepath.suffix == '.vg'
        assert self.planes is not None and self.points_grouped is not None

        # points
        out = ''
        out += 'num_points: {}\n'.format(len(self.points))
        for i in self.points.flatten():
            # https://stackoverflow.com/questions/54367816/numpy-np-fromstring-not-working-as-hoped
            out += '{} '.format(i)

        # colors (no color needed)
        out += '\nnum_colors: {}'.format(0)

        # normals
        out += '\nnum_normals: {}\n'.format(len(self.points))
        if self.normals is None:
            for i, group in enumerate(self.points_grouped):
                for _ in group:
                    out += '{} {} {} '.format(*self.planes[i][:3])
            num_remainder_points = len(self.points) - sum(len(g) for g in self.points_grouped)
            for _ in range(num_remainder_points):
                out += '{} {} {} '.format(random(), random(), random())
        else:
            for n in self.normals:
                out += '{} {} {} '.format(*n)

        # groups
        out += '\nnum_groups: {}\n'.format(len(self.points_grouped))
        j_base = 0
        for i, group in enumerate(self.points_grouped):
            out += 'group_type: {}\n'.format(0)
            out += 'num_group_parameters: {}\n'.format(4)
            out += 'group_parameters: {} {} {} {}\n'.format(*self.planes[i])
            out += 'group_label: group_{}\n'.format(i)
            out += 'group_color: {} {} {}\n'.format(random(), random(), random())
            out += 'group_num_point: {}\n'.format(len(self.points_grouped[i]))
            for j in range(j_base, j_base + len(self.points_grouped[i])):
                out += '{} '.format(j)
            j_base += len(self.points_grouped[i])
            out += '\nnum_children: {}\n'.format(0)

        with open(filepath, 'w') as fout:
            fout.writelines(out)

    def save_bvg(self, filepath):
        """
        Save primitives into a bvg file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save bvg file
        """

        if isinstance(filepath, str):
            assert filepath.endswith('.bvg')
        elif isinstance(filepath, Path):
            assert filepath.suffix == '.bvg'
        assert self.planes is not None and self.points_grouped is not None

        # points
        out = [struct.pack('i', len(self.points))]
        for i in self.points.flatten():
            # https://stackoverflow.com/questions/54367816/numpy-np-fromstring-not-working-as-hoped
            out.append(struct.pack('f', i))

        # colors (no color needed)
        out.append(struct.pack('i', 0))

        # normals
        out.append(struct.pack('i', len(self.points)))
        if self.normals is None:
            for i, group in enumerate(self.points_grouped):
                for _ in group:
                    out.append(struct.pack('fff', *self.planes[i][:3]))
            num_remainder_points = len(self.points) - sum(len(g) for g in self.points_grouped)
            for _ in range(num_remainder_points):
                out.append(struct.pack('fff', random(), random(), random()))
        else:
            for n in self.normals:
                out.append(struct.pack('fff', *n))

        # groups
        out.append(struct.pack('i', len(self.points_grouped)))
        j_base = 0
        for i, group in enumerate(self.points_grouped):
            out.append(struct.pack('i', 0))
            out.append(struct.pack('i', 4))
            out.append(struct.pack('ffff', *self.planes[i]))
            out.append(struct.pack('i', 6 + len(str(i))))
            out.append(struct.pack(f'{(6 + len(str(i)))}s', bytes('group_{}'.format(i), encoding='ascii')))
            out.append(struct.pack('fff', random(), random(), random()))
            out.append(struct.pack('i', len(self.points_grouped[i])))

            for j in range(j_base, j_base + len(self.points_grouped[i])):
                out.append(struct.pack('i', j))

            j_base += len(self.points_grouped[i])
            out.append(struct.pack('i', 0))

        with open(filepath, 'wb') as fout:
            fout.writelines(out)

    def save_cloud(self, filepath):
        """
        Save point cloud into a common 3D format. Support formats: xyzn, xyzrgb, pts, ply, pcd.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save point cloud file
        """
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(self.points)
        if self.normals is not None:
            pcd.normals = o3d.utility.Vector3dVector(self.normals)
        o3d.io.write_point_cloud(str(filepath), pcd)



# Copyright (c) 2022 Zhaiyu Chen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
complex.py
----------

Cell complex from planar primitive arrangement.

A linear cell complex is constructed from planar primitives
with adaptive binary space partitioning: upon insertion of a primitive
only the local cells that are intersecting it will be updated,
so will be the corresponding adjacency graph of the complex.
"""

import os
import string
from pathlib import Path
import itertools
import heapq
from copy import copy
from random import random, choices, uniform, randint
import pickle
import time
import multiprocessing

import numpy as np
from tqdm import trange
import networkx as nx
import trimesh
from sage.all import polytopes, QQ, RR, Polyhedron
from sage.geometry.triangulation.point_configuration import PointConfiguration
from sklearn.decomposition import PCA

from .logger import attach_to_log
from .primitive import VertexGroup

logger = attach_to_log()


class CellComplex:
    """
    Class of cell complex from planar primitive arrangement.
    """

    def __init__(self, planes, aabbs, obbs=None, points=None, initial_bound=None, initial_padding=0.1, additional_planes=None,
                 build_graph=False, quiet=False):
        """
        Init CellComplex.
        Class of cell complex from planar primitive arrangement.

        Parameters
        ----------
        planes: (n, 4) float
            Plana parameters
        aabbs: (n, 2, 3) float
            Corresponding bounding box bounds of the planar primitives
        obbs: (n, 4, 3) float
            Corresponding oriented bounds of the planar primitives
        points: (n, ) object of float
            Points grouped into primitives, points[any]: (m, 3)
        initial_bound: None or (2, 3) float
            Initial bound to partition
        build_graph: bool
            Build the cell adjacency graph if set True.
        additional_planes: None or (n, 4) float
            Additional planes to append to the complex,
            can be missing planes due to occlusion or incapacity of RANSAC
        quiet: bool
            Disable logging and progress bar if set True
        """
        self.quiet = quiet
        if self.quiet:
            logger.disabled = True

        self.aabbs = aabbs  # numpy.array over RDF
        self.obbs = obbs      # numpy.array over RDF
        self.planes = planes  # numpy.array over RDF
        self.points = points
        # missing planes due to occlusion or incapacity of RANSAC
        self.additional_planes = additional_planes

        self.initial_bound = initial_bound if initial_bound else self._pad_bound(
            [np.amin(aabbs[:, 0, :], axis=0), np.amax(aabbs[:, 1, :], axis=0)],
            padding=initial_padding)
        self.cells = [self._construct_initial_cell()]  # list of QQ
        self.cells_bounds = [self.cells[0].bounding_box()]  # list of QQ

        if build_graph:
            self.graph = nx.Graph()
            self.graph.add_node(0)  # the initial cell
            self.index_node = 0  # unique for every cell ever generated
        else:
            self.graph = None

        self.constructed = False

    def _construct_initial_cell(self):
        """
        Construct initial bounding cell.

        Return
        ----------
        as_object: Polyhedron object
            Polyhedron object of the initial cell,
            a cuboid with 12 triangular facets.
        """
        return polytopes.cube(
            intervals=[[QQ(self.initial_bound[0][i]), QQ(self.initial_bound[1][i])] for i in range(3)])

    def refine_planes(self, theta=10 * 3.1416 / 180, epsilon=0.005, normalise_normal=False):
        """
        Refine planar primitives.

        First, compute the angle of the supporting planes for each pair of planar primitives.
        Then, starting from the pair with the smallest angle, test if the following two conditions are met:
        (a) the angle between is lower than a threshold. (b) more than a specified number of points lie on
        both primitives. Merge the two primitives and fit a new plane if the conditions are satisfied.

        Parameters
        ----------
        theta: float
            Angle tolerance, primitive pair has to be less than this tolerance to be refined
        epsilon: float
            Distance tolerance, primitive pair has to be less than this tolerance to be refined
        normalise_normal: bool
            Normalise normal if set True
        """
        if self.points is None:
            raise ValueError('point coordinates are needed for plane refinement')
        logger.info('refining planar primitives')
        if self.additional_planes:
            logger.warning('additional planes are not refined')

        # shallow copy of the primitives
        planes = list(copy(self.planes))
        aabbs = list(copy(self.aabbs))
        obbs = list(copy(self.obbs))
        points = list(copy(self.points))

        # pre-compute cosine of theta
        theta_cos = np.cos(theta)

        # priority queue storing pairwise planes and their angles
        priority_queue = []

        # compute angles and save them to the priority queue
        for i, j in itertools.combinations(range(len(planes)), 2):
            # no need to normalise as PCA in primitive.py already does it
            angle_cos = np.abs(np.dot(planes[i][:3], planes[j][:3]))
            if normalise_normal:
                angle_cos /= (np.linalg.norm(planes[i][:3]) * np.linalg.norm(planes[j][:3]))
            heapq.heappush(priority_queue, [-angle_cos, i, j])  # negate to use max-heap

        # indices of planes to be merged
        to_merge = set()

        while priority_queue:
            # pair with the smallest angle
            pair = heapq.heappop(priority_queue)

            if -pair[0] > theta_cos:  # negate back to use max-heap
                # distance from the center of a primitive to the supporting plane of the other
                distance = np.abs(
                    np.dot((np.array(points[pair[1]]).mean(axis=0) - np.array(points[pair[2]]).mean(axis=0)),
                           planes[pair[1]][:3]))

                # assume normalised data so that epsilon can be an arbitrary number in (0, 1]
                if distance < epsilon and pair[1] not in to_merge and pair[2] not in to_merge:
                    # merge the two planes
                    points_merged = np.concatenate([points[pair[1]], points[pair[2]]])
                    planes_merged, obbs_merged = VertexGroup.fit_plane(points_merged)
                    aabbs_merged = [np.min([aabbs[pair[1]][0], aabbs[pair[2]][0]], axis=0).tolist(),
                                    np.max([aabbs[pair[1]][1], aabbs[pair[2]][1]], axis=0).tolist()]

                    # update to_merge
                    to_merge.update({pair[1]})
                    to_merge.update({pair[2]})

                    # push the merged plane into the heap
                    for i, p in enumerate(planes):
                        if i not in to_merge:
                            angle_cos = np.abs(np.dot(planes_merged[:3], p[:3]))
                            heapq.heappush(priority_queue, [-angle_cos, i, len(planes)])  # placeholder

                    # update the actual data with merged ones
                    points.append(points_merged)
                    planes.append(planes_merged)
                    aabbs.append(aabbs_merged)
                    obbs.append(obbs_merged)

            else:
                # no more possible coplanar pairs can exist in this priority queue
                break

        # delete the merged pairs
        for i in sorted(to_merge, reverse=True):
            del planes[i]
            del aabbs[i]
            del obbs[i]
            del points[i]

        logger.info('{} pairs of planes merged'.format(len(to_merge)))

        self.planes = np.array(planes)
        self.aabbs = np.array(aabbs)
        self.obbs = np.array(obbs)
        self.points = np.array(points, dtype=object)

    def prioritise_planes(self, prioritise_verticals=True):
        """
        Prioritise certain planes to favour building reconstruction.

        First, vertical planar primitives are accorded higher priority than horizontal or oblique ones
        to avoid incomplete partitioning due to missing data about building facades.
        Second, in the same priority class, planar primitives with larger areas are assigned higher priority
        than smaller ones, to make the final cell complex as compact as possible.
        Note that this priority setting is designed exclusively for building models.

        Parameters
        ----------
        prioritise_verticals: bool
            Prioritise vertical planes if set True
        """
        logger.info('prioritising planar primitives')
        # compute the priority
        indices_sorted_planes = self._sort_planes()

        if prioritise_verticals:
            indices_vertical_planes = self._vertical_planes(slope_threshold=0.9)
            bool_vertical_planes = np.in1d(indices_sorted_planes, indices_vertical_planes)
            indices_priority = np.append(indices_sorted_planes[bool_vertical_planes],
                                         indices_sorted_planes[np.invert(bool_vertical_planes)])
        else:
            indices_priority = indices_sorted_planes

        # reorder both the planes and their bounds
        self.planes = self.planes[indices_priority]
        self.aabbs = self.aabbs[indices_priority]

        # reorder obbs and points if not None
        if self.obbs is not None:
            self.obbs = self.obbs[indices_priority]
        if self.points is not None:
            self.points = self.points[indices_priority]

        # append additional planes with the highest priority
        if self.additional_planes:
            self.planes = np.concatenate([self.additional_planes, self.planes], axis=0)
            additional_bounds = [[[-np.inf, -np.inf, -np.inf], [np.inf, np.inf, np.inf]]] * len(self.additional_planes)
            self.aabbs = np.concatenate([additional_bounds, self.aabbs], axis=0)  # never miss an intersection
            if self.obbs is not None:
                additional_obbs = [[[-np.inf, -np.inf, -np.inf], [-np.inf, np.inf, -np.inf], [np.inf, np.inf, -np.inf],
                                    [np.inf, -np.inf, -np.inf]]] * len(self.additional_planes)
                self.obbs = np.concatenate([additional_obbs, self.obbs], axis=0)
            if self.points is not None:
                additional_points = np.empty((1)) * len(self.additional_planes)
                self.points = np.concatenate([additional_points, self.points], axis=0)

        logger.debug('ordered planes: {}'.format(self.planes))
        logger.debug('ordered aabbs: {}'.format(self.aabbs))
        logger.debug('ordered obbs: {}'.format(self.obbs))

    def _vertical_planes(self, slope_threshold=0.9, epsilon=10e-5):
        """
        Return vertical planes.

        Parameters
        ----------
        slope_threshold: float
            Slope threshold, above which the planes are considered vertical
        epsilon: float
            Trivial term to avoid NaN

        Returns
        -------
        as_int: (n,) int
            Indices of the vertical planar primitives
        """
        slope_squared = (self.planes[:, 0] ** 2 + self.planes[:, 1] ** 2) / (self.planes[:, 2] ** 2 + epsilon)
        return np.where(slope_squared > slope_threshold ** 2)[0]

    def _sort_planes(self, mode='norm'):
        """
        Sort planes.

        Parameters
        ----------
        mode: str
            Mode for sorting, can be 'volume' or 'norm'

        Returns
        -------
        as_int: (n,) int
            Indices by which the planar primitives are sorted based on their bounding box volume
        """
        if mode == 'volume':
            volume = np.prod(self.aabbs[:, 1, :] - self.aabbs[:, 0, :], axis=1)
            return np.argsort(volume)[::-1]
        elif mode == 'norm':
            sizes = np.linalg.norm(self.aabbs[:, 1, :] - self.aabbs[:, 0, :], ord=2, axis=1)
            return np.argsort(sizes)[::-1]
        elif mode == 'area':
            # project the points supporting each plane onto the plane
            # https://stackoverflow.com/questions/9605556/how-to-project-a-point-onto-a-plane-in-3d
            raise NotImplementedError
        else:
            raise ValueError('mode has to be "volume" or "norm"')

    @staticmethod
    def _pad_bound(bound, padding=0.00):
        """
        Pad bound.

        Parameters
        ----------
        bound: (2, 3) float
            Bound of the query planar primitive
        padding: float
            Padding factor, defaults to 0.05.

        Returns
        -------
        as_float: (2, 3) float
            Padded bound
        """
        extent = bound[1] - bound[0]
        return [bound[0] - extent * padding, bound[1] + extent * padding]

    def _intersect_bound_plane(self, bound, plane, exhaustive=False, epsilon=10e-8):
        """
        Pre-intersection test between query primitive and existing cells, based on AABBs and plane parameters.

        Parameters
        ----------
        bound: (2, 3) float
            Bound of the query planar primitive
        plane: (4,) float
            Plane parameters
        exhaustive: bool
            Exhaustive partitioning, only for benchmarking
        epsilon: float
            Distance tolerance

        Returns
        -------
        as_int: (n,) int
            Indices of existing cells whose bounds intersect with bounds of the query primitive
            and intersect with the supporting plane of the primitive
        """
        if exhaustive:
            return np.arange(len(self.cells_bounds))

        # each planar primitive partitions only the 3D cells that intersect with it
        cells_bounds = np.array(self.cells_bounds)  # easier array manipulation
        center_targets = np.mean(cells_bounds, axis=1)  # N * 3
        extent_targets = cells_bounds[:, 1, :] - cells_bounds[:, 0, :]  # N * 3

        if bound[0][0] == -np.inf:
            intersection_bound = np.arange(len(self.cells_bounds))

        else:
            # intersection with existing cells' AABB
            center_query = np.mean(bound, axis=0)  # 3,
            center_distance = np.abs(center_query - center_targets)  # N * 3
            extent_query = bound[1] - bound[0]  # 3,

            # abs(center_distance) * 2 < (query extent + target extent) for every dimension -> intersection
            intersection_bound = \
                np.where(np.all(center_distance * 2 < extent_query + extent_targets + epsilon, axis=1))[0]

        # plane-AABB intersection test from extracted intersection_bound only
        # https://gdbooks.gitbooks.io/3dcollisions/content/Chapter2/static_aabb_plane.html
        # compute the projection interval radius of AABB onto L(t) = center + t * normal
        radius = np.dot(extent_targets[intersection_bound] / 2, np.abs(plane[:3]))
        # compute distance of box center from plane
        distance = np.dot(center_targets[intersection_bound], plane[:3]) + plane[3]
        # intersection between plane and AABB occurs when distance falls within [-radius, +radius] interval
        intersection_plane = np.where(np.abs(distance) <= radius + epsilon)[0]

        return intersection_bound[intersection_plane]

    def _intersect_obb(self, plane, obb, indices_cells, exhaustive=False, epsilon=10e-8):
        """
        Further-intersection test between query primitive and existing cells, based on their OBBs.

        Parameters
        ----------
        plane: (4,) float
            Plane parameters
        obb: (4, 3) float
            Oriented bounds of the planar primitive
        indices_cells: (n,) int
            Indices of existing cells whose bounds intersect with bounds of the query primitive
            and intersect with the supporting plane of the primitive in the first intersection test
        exhaustive: bool
            Exhaustive partitioning, only for benchmarking
        epsilon: float
            Distance tolerance

        Returns
        -------
        as_int: (n,) int
            Indices of existing cells whose bounds intersect with bounds of the query primitive
            and intersect with the supporting plane of the primitive
        """
        if exhaustive:
            return indices_cells

        vertices_query = obb  # 4 * 3
        norm_query = self._normalize(plane[:3])  # 3,
        edges_query = np.array([obb[0, :] - obb[1, :], obb[1, :] - obb[2, :]])  # 2 * 3
        edges_query = self._normalize(edges_query)  # 2 * 3

        vertices_targets = [cell.vertices_list() for i, cell in enumerate(self.cells) if i in indices_cells]
        norms_targets = [self._cell_norm(vertices) for vertices in vertices_targets]
        edges_targets = np.array(norms_targets)  # N * 3 * 3

        # all vertex lists have the same length for parallelization
        max_vertices = max([len(vertices) for vertices in vertices_targets])
        vertices_targets = [vertices + [vertices[-1]] * (max_vertices - len(vertices)) for vertices in vertices_targets]
        vertices_targets = np.array(vertices_targets)  # N * M * 3

        # 4 possible separating axis
        # 1: plane normal
        axis = norm_query  # 3,
        projection_query = np.einsum('ni,i->n', vertices_query, axis)  # 4,
        min_projection_query = np.min(projection_query, axis=0)  # 1,
        max_projection_query = np.max(projection_query, axis=0)  # 1,
        projection_targets = np.einsum('nmi,i->nm', vertices_targets, axis)
        min_projection_targets = np.min(projection_targets, axis=1)  # m,
        max_projection_targets = np.max(projection_targets, axis=1)  # m,
        condition1 = (min_projection_query > max_projection_targets + epsilon) | (
                min_projection_targets > max_projection_query + epsilon)

        # 2: cell_obb normals
        axis = norms_targets  # N * 3 * 3
        projection_query = np.einsum('nmi,bi->nmb', axis, vertices_query)  # N * 3 * 4
        min_projection_query = np.min(projection_query, axis=2)  # N * 3
        max_projection_query = np.max(projection_query, axis=2)  # N * 3
        projection_targets = np.einsum('nmi,nbi->nmb', axis, vertices_targets)  # N * 3 * M
        min_projection_targets = np.min(projection_targets, axis=2)  # N * 3 
        max_projection_targets = np.max(projection_targets, axis=2)  # N * 3
        condition2 = (min_projection_query > max_projection_targets + epsilon) | (
                min_projection_targets > max_projection_query + epsilon)
        condition2 = np.any(condition2, axis=1)

        # 3: plane normal cross plane_obb edges
        axis = np.cross(norm_query, edges_query)  # 2 * 3
        axis = self._normalize(axis)  # 2 * 3
        projection_query = np.einsum('ni,mi->nm', axis, vertices_query)  # 2 * 4
        min_projection_query = np.min(projection_query, axis=1)  # 2 * 3
        max_projection_query = np.max(projection_query, axis=1)  # 2 * 3
        projection_targets = np.einsum('ni,mbi->mnb', axis, vertices_targets)
        min_projection_targets = np.min(projection_targets, axis=2)
        max_projection_targets = np.max(projection_targets, axis=2)  # m * 2 * 3
        condition3 = (min_projection_query > max_projection_targets + epsilon) | (
                min_projection_targets > max_projection_query + epsilon)
        condition3 = np.any(condition3, axis=1)

        # 4: cell_obb edges cross plane_obb edges
        axis = np.cross(edges_query[np.newaxis, np.newaxis, :, :], edges_targets[:, :, np.newaxis, :])  # N * 3 * 2 * 3
        axis = axis.reshape(-1, 6, 3)  # N * 6 * 3
        axis = self._normalize(axis)  # N * 6 * 3
        projection_query = np.einsum('nmi,bi->nmb', axis, vertices_query)  # N * 6 * 4
        min_projection_query = np.min(projection_query, axis=2)  # N * 6
        max_projection_query = np.max(projection_query, axis=2)  # N * 6
        projection_targets = np.einsum('mni,mbi->mnb', axis, vertices_targets)  # N * 6 * M
        min_projection_targets = np.min(projection_targets, axis=2)  # N * 6
        max_projection_targets = np.max(projection_targets, axis=2)  # N * 6
        condition4 = (min_projection_query > max_projection_targets + epsilon) | (
                min_projection_targets > max_projection_query + epsilon)
        condition4 = np.any(condition4, axis=1)

        # combine all conditions
        condition = condition1 | condition2 | condition3 | condition4
        condition = np.array([not item for item in condition])
        intersection_plane = np.where(condition)[0]
        intersection_plane = indices_cells[intersection_plane]

        # if the OBB is unbounded, the intersection is incomplete
        if obb[0][0] == -np.inf and np.any(indices_cells == 0) and not np.any(intersection_plane == 0):
            intersection_plane = np.append(intersection_plane, 0)

        return intersection_plane

    @staticmethod
    def _normalize(axis):
        """
        Normalize axis.

        Parameters
        ----------
        axis: (n_0, n_1, ..., n_k, 3) float
        
        Returns
        -------
        as_float: (n_0, n_1, ..., n_k, 3) float
            Normalized axis
        """
        eps = 1e-12
        enom = np.linalg.norm(axis, axis=-1, keepdims=True)
        enom_clipped = np.clip(enom, eps, None)
        return axis / enom_clipped

    @staticmethod
    def _cell_norm(vertices):
        """
        Compute normals of a cell's oriented bounding box.

        Parameters
        ----------
        vertices: (n, 3) float
            cell's vertices list

        Returns
        -------        
        normals: (3,3) float
            Normals of the cell's oriented bounding box
        """
        pca = PCA(n_components=3)
        pca.fit(vertices)
        normals = pca.components_
        return normals

    @staticmethod
    def _inequalities(plane):
        """
        Inequalities from plane parameters.

        Parameters
        ----------
        plane: (4,) float
            Plane parameters

        Returns
        -------
        positive: (4,) float
            Inequality of the positive half-plane
        negative: (4,) float
            Inequality of the negative half-plane
        """
        positive = [QQ(plane[-1]), QQ(plane[0]), QQ(plane[1]), QQ(plane[2])]
        negative = [QQ(-element) for element in positive]
        return positive, negative

    def _index_node_to_cell(self, query):
        """
        Convert index in the node list to that in the cell list.
        The rationale behind is #nodes == #cells (when a primitive is settled down).

        Parameters
        ----------
        query: int
            Query index in the node list

        Returns
        -------
        as_int: int
            Query index in the cell list
        """
        return list(self.graph.nodes).index(query)

    def _intersect_neighbour(self, kwargs):
        """
        Intersection test between partitioned cells and neighbouring cell.
        Implemented for multi-processing across all neighbours.

        Parameters
        ----------
        kwargs: (int, Polyhedron object, Polyhedron object, Polyhedron object)
            (neighbour index, positive cell, negative cell, neighbouring cell)
        """
        n, cell_positive, cell_negative, cell_neighbour = kwargs['n'], kwargs['positive'], kwargs['negative'], kwargs[
            'neighbour']

        interface_positive = cell_positive.intersection(cell_neighbour)

        if interface_positive.dim() == 2:
            # this neighbour can connect with either or both children
            self.graph.add_edge(self.index_node + 1, n)
            interface_negative = cell_negative.intersection(cell_neighbour)
            if interface_negative.dim() == 2:
                self.graph.add_edge(self.index_node + 2, n)
        else:
            # this neighbour must otherwise connect with the other child
            self.graph.add_edge(self.index_node + 2, n)

    def construct(self, exhaustive=False, num_workers=0):
        """
        Construct cell complex.

        Two-stage primitive-in-cell predicate.
        (1) bounding boxes of primitive and existing cells are evaluated
        for possible intersection. (2), a strict intersection test is performed.

        Generated cells are stored in self.cells.
        * query the bounding box intersection.
        * optional: intersection test for polygon and edge in each potential cell.
        * partition the potential cell into two. rewind if partition fails.

        Parameters
        ----------
        exhaustive: bool
            Do exhaustive partitioning if set True
        
        num_workers: int
            Number of workers for multi-processing, disabled if set 0
        """
        logger.info('constructing cell complex')
        tik = time.time()

        pool = None
        if num_workers > 0:
            pool = multiprocessing.Pool(processes=num_workers)

        pbar = range(len(self.aabbs)) if self.quiet else trange(len(self.aabbs))
        for i in pbar:  # kinetic for each primitive
            # bounding box intersection test
            # indices of existing cells with potential intersections
            indices_cells = self._intersect_bound_plane(self.aabbs[i], self.planes[i], exhaustive)

            if self.obbs is not None:
                indices_cells = self._intersect_obb(self.planes[i], self.obbs[i], indices_cells, exhaustive)

            assert len(indices_cells), 'intersection failed! check the initial bound'

            # half-spaces defined by inequalities
            # no change_ring() here (instead, QQ() in _inequalities) speeds up 10x
            # init before the loop could possibly speed up a bit
            hspace_positive, hspace_negative = [Polyhedron(ieqs=[inequality]) for inequality in
                                                self._inequalities(self.planes[i])]

            # partition the intersected cells and their bounds while doing mesh slice plane
            indices_parents = []

            for index_cell in indices_cells:
                cell_positive = hspace_positive.intersection(self.cells[index_cell])
                cell_negative = hspace_negative.intersection(self.cells[index_cell])

                if cell_positive.dim() != 3 or cell_negative.dim() != 3:
                    # if cell_positive.is_empty() or cell_negative.is_empty():
                    """
                    cannot use is_empty() predicate for degenerate cases:
                        sage: Polyhedron(vertices=[[0, 1, 2]])
                        A 0-dimensional polyhedron in ZZ^3 defined as the convex hull of 1 vertex
                        sage: Polyhedron(vertices=[[0, 1, 2]]).is_empty()
                        False
                    """
                    continue

                # incrementally build the adjacency graph
                if self.graph is not None:
                    # append the two nodes (UID) being partitioned
                    self.graph.add_node(self.index_node + 1)
                    self.graph.add_node(self.index_node + 2)

                    # append the edge in between
                    self.graph.add_edge(self.index_node + 1, self.index_node + 2)

                    # get neighbours of the current cell from the graph
                    neighbours = self.graph[list(self.graph.nodes)[index_cell]]  # index in the node list

                    if neighbours:
                        # get the neighbouring cells to the parent
                        cells_neighbours = [self.cells[self._index_node_to_cell(n)] for n in neighbours]

                        # adjacency test between both created cells and their neighbours
                        # todo:
                        #   Avoid 3d-3d intersection if possible. Unsliced neighbours connect with only one child;
                        #   sliced neighbors connect with both children.

                        kwargs = []
                        for n, cell in zip(neighbours, cells_neighbours):
                            kwargs.append(
                                {'n': n, 'positive': cell_positive, 'negative': cell_negative, 'neighbour': cell})

                        if pool is None:
                            for k in kwargs:
                                self._intersect_neighbour(k)
                        else:
                            pool.map(self._intersect_neighbour, kwargs)

                    # update cell id
                    self.index_node += 2

                self.cells.append(cell_positive)
                self.cells.append(cell_negative)

                # incrementally cache the bounds for created cells
                self.cells_bounds.append(cell_positive.bounding_box())
                self.cells_bounds.append(cell_negative.bounding_box())

                indices_parents.append(index_cell)

            # delete the parent cells and their bounds. this does not affect the appended ones
            for index_parent in sorted(indices_parents, reverse=True):
                del self.cells[index_parent]
                del self.cells_bounds[index_parent]

                # remove the parent node (and subsequently its incident edges) in the graph
                if self.graph is not None:
                    self.graph.remove_node(list(self.graph.nodes)[index_parent])

        self.constructed = True
        logger.info('cell complex constructed: {:.2f} s'.format(time.time() - tik))

    def visualise(self, indices_cells=None, temp_dir='./'):
        """
        Visualise the cells using trimesh.

        pyglet installation is needed for the visualisation.

        Parameters
        ----------
        indices_cells: None or (n,) int
            Indices of cells to be visualised
        temp_dir: str or Path
            Temp dir to save intermediate visualisation
        """
        if self.constructed:
            try:
                import pyglet
            except ImportError:
                logger.warning('pyglet installation not found; skip visualisation')
                return
            if indices_cells is not None and len(indices_cells) == 0:
                raise ValueError('no indices provided')

            filename_stem = ''.join(choices(string.ascii_uppercase + string.digits, k=5))
            filename_mesh = filename_stem + '.obj'
            filename_mtl = filename_stem + '.mtl'

            self.save_obj(filepath=temp_dir + filename_mesh, indices_cells=indices_cells, use_mtl=True)
            scene = trimesh.load_mesh(temp_dir + filename_mesh)
            scene.show()
            os.remove(temp_dir + filename_mesh)
            os.remove(temp_dir + filename_mtl)
        else:
            raise RuntimeError('cell complex has not been constructed')

    @property
    def num_cells(self):
        """
        Number of cells in the complex.
        """
        return len(self.cells)

    @property
    def num_planes(self):
        """
        Number of planes in the complex, excluding the initial bounding box.
        """
        return len(self.planes)

    def volumes(self, multiplier=1.0, engine='Qhull'):
        """
        list of cell volumes.

        Parameters
        ----------
        multiplier: float
            Multiplier to the volume
        engine: str
            Engine to compute volumes, can be 'Qhull' or 'Sage' with native SageMath

        Returns
        -------
        as_float: list of float
            Volumes of cells
        """
        if engine == 'Qhull':
            from scipy.spatial import ConvexHull
            volumes = [None for _ in range(len(self.cells))]
            for i, cell in enumerate(self.cells):
                try:
                    volumes[i] = ConvexHull(cell.vertices_list()).volume * multiplier
                except:
                    # degenerate floating-point
                    volumes[i] = RR(cell.volume()) * multiplier
            return volumes

        elif engine == 'Sage':
            return [RR(cell.volume()) * multiplier for cell in self.cells]

        else:
            raise ValueError('engine must be either "Qhull" or "Sage"')

    def cell_representatives(self, location='center', num=1):
        """
        Return representatives of cells in the complex.

        Parameters
        ----------
        location: str
            'center' represents the average of the vertices of the polyhedron,
            'centroid' represents the center of mass/volume,
            'random_r' represents random point(s) by rejection,
            'random_t' represents random point(s) by tetrahedralization,
            'skeleton' represents skeleton point(s),
            'boundary' represents boundary point(s) by triangulation
        num: int
            number of samples per cell, only applies to 'random' and 'star'

        Returns
        -------
        as_float: (n, 3) float for 'center' and 'centroid', or (m, n, 3) for 'random' and 'skeleton' and 'boundary'
            Representatives of cells in the complex.
        """
        points = []

        if location == 'center':
            return [cell.center() for cell in self.cells]

        elif location == 'centroid':
            return [cell.centroid() for cell in self.cells]

        elif location == 'random_r':
            # strict random by sampling and rejection
            for cell in self.cells:
                bbox = cell.bounding_box()
                points_cell = []
                while len(points_cell) < num:
                    sample = (uniform(bbox[0][0], bbox[1][0]), uniform(bbox[0][1], bbox[1][1]),
                              uniform(bbox[0][2], bbox[1][2]))
                    if cell.contains(sample):
                        points_cell.append(sample)
                points.append(points_cell)
            return points

        elif location == 'random_t':
            # strict random by triangulation and sampling
            def tetrahedron_volume(a, b, c, d, epsilon=10e-6):
                ab = b - a
                ac = c - a
                ad = d - a
                v = abs(np.dot(ab, np.cross(ac, ad))) / 6.0 + epsilon  # epsilon here to prevent empty volume
                return v

            for cell in self.cells:
                points_cell = []

                # triangulate cell
                vertices = cell.vertices()
                point_config = PointConfiguration(vertices)
                triangulation = point_config.triangulate()
                vertices = np.array([[vertices[v] for v in t] for t in triangulation])
                volumes = [tetrahedron_volume(*tetra) for tetra in vertices]
                options = list(range(len(volumes)))

                for _ in range(num):
                    # select one tetrahedron with volume-based probability
                    choice = choices(options, volumes)[0]

                    # compute vertex-based probability
                    u, v, w = sorted([random() for _ in range(3)])
                    u, v, w = u, v - u, w - v

                    # randomly sample one point within the tetrahedron
                    v1, v2, v3, v4 = vertices[choice]
                    point = u * v1 + v * v2 + w * v3 + (1 - u - v - w) * v4
                    points_cell.append(point)
                points.append(points_cell)
            return points

        elif location == 'skeleton':
            # star-shaped sampling
            for cell in self.cells:
                vertices = cell.vertices_list()
                if num <= len(vertices):
                    # vertices given high priority
                    points.append(choices(vertices, k=num))
                else:
                    num_per_vertex = num // len(vertices)
                    num_remainder = num % len(vertices)
                    centroid = cell.centroid()
                    points_cell = []
                    for vertex in vertices[:-1]:
                        points_cell.extend([vertex + (centroid - np.array(vertex)) / num_per_vertex * i
                                            for i in range(num_per_vertex)])
                    # last vertex consumes remainder points
                    points_cell.extend(
                        [vertices[-1] + (centroid - np.array(vertices[-1])) / (num_remainder + num_per_vertex)
                         * i for i in range(num_remainder + num_per_vertex)])
                    points.append(points_cell)
            return points

        elif location == 'boundary':
            # boundary sampling
            def triangle_area(a, b, c, epsilon=10e-6):
                ab = b - a
                ac = c - a
                cross_prod = np.cross(ab, ac)
                area = 0.5 * np.linalg.norm(cross_prod) + epsilon
                return area

            for cell in self.cells:
                points_cell = []

                # get the list of facets
                facets = cell.facets()

                # aggregate all triangles
                triangles = []
                for f in facets:
                    vertices = f.vertices()
                    point_config = PointConfiguration(vertices)
                    triangulation = point_config.triangulate()  # degrade to 2D triangulation
                    vertices = np.array([[vertices[v] for v in t] for t in triangulation])
                    triangles.extend(vertices)

                # assign triangle-wise probabilities proportional to their areas
                areas = [triangle_area(*triangle) for triangle in triangles]
                options = list(range(len(areas)))

                for _ in range(num):
                    # select one triangle with area-based probability
                    choice = choices(options, areas)[0]

                    # compute vertex-based probability
                    u, v = sorted([random() for _ in range(2)])
                    u, v = u, v - u

                    # randomly sample one point within the triangle
                    v1, v2, v3 = triangles[choice]
                    point = u * v1 + v * v2 + (1 - u - v) * v3
                    points_cell.append(point)
                points.append(points_cell)
            return points

        else:
            raise ValueError(
                "expected 'center', 'centroid', 'random_r', 'random_t', 'skeleton' or 'boundary' as mode, got {}".format(
                    location))

    def cells_boundary(self, epsilon=1e-5):
        """
        Return indices of boundary cells (touching the bounds).
        These cells could be excluded from a valid reconstruction, only when a significant padding exists.

        Parameters
        ----------
        epsilon: float
            Distance tolerance

        Returns
        -------
        as_int: (n, ) int
            Indices of boundary cells
        """
        initial_bound = np.array(self.initial_bound)
        indices = []
        for index, cell_bound in enumerate(self.cells_bounds):
            if (cell_bound[0] < np.array(initial_bound[0]) + epsilon).any() \
                    or (cell_bound[1] > initial_bound[1] - epsilon).any():
                indices.append(index)
        return indices

    def cells_in_mesh(self, filepath_mesh, engine='distance'):
        """
        Return indices of cells that are inside a reference mesh.

        Parameters
        ----------
        filepath_mesh: str or Path
            Filepath to reference mesh
        engine: str
            Engine to compute predicate, can be 'embree' for pyembree (up to x50 faster but not stable),
            'ray_triangle' for a slower ray tracer, or 'distance' for signed distance

        Returns
        -------
        as_int: (n, ) int
            Indices of cells being inside the reference mesh
        """
        mesh = trimesh.load_mesh(filepath_mesh)
        centers = self.cell_representatives(location='center')

        if engine == 'embree':
            # not stable, and no exact returns, use with caution
            # https://trimsh.org/trimesh.ray.ray_pyembree.html
            # https://github.com/mikedh/trimesh/issues/242
            # https://github.com/mikedh/trimesh/issues/306
            logger.warning('ray tracing unstable; use with caution')
            contains = trimesh.ray.ray_pyembree.RayMeshIntersector(mesh).contains_points(centers)
            return contains.nonzero()[0]
        
        elif engine == 'ray_triangle':
            # not stable, use with caution
            # https://trimsh.org/trimesh.ray.ray_triangle.html
            # https://github.com/mikedh/trimesh/issues/242
            logger.warning('ray tracing unstable; use with caution')
            contains = mesh.contains(centers)
            return contains.nonzero()[0]

        elif engine == 'distance':
            # https://trimsh.org/trimesh.proximity.html
            distances = trimesh.proximity.signed_distance(mesh, centers)
            return (distances >= 0).nonzero()[0]
        else:
            raise ValueError("expected 'embree', 'ray_triangle' or 'distance' as engine, got {}".format(engine))

    def print_info(self):
        """
        Print info to console.
        """
        logger.info('number of planes: {}'.format(self.num_planes))
        logger.info('number of cells: {}'.format(self.num_cells))

    def save(self, filepath):
        """
        Save the cell complex to a CC file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save CC file, '.cc' suffix recommended
        """
        if self.constructed:
            # create the dir if not exists
            with open(filepath, 'wb') as f:
                pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
        else:
            raise RuntimeError('cell complex has not been constructed')

    def save_npy(self, filepath):
        """
        Save the cells to an npy file (deprecated).

        Parameters
        ----------
        filepath: str or Path
            Filepath to save npy file
        """
        if self.constructed:
            # create the dir if not exists
            filepath = Path(filepath)
            filepath.parent.mkdir(parents=True, exist_ok=True)
            np.save(filepath, self.cells, allow_pickle=True)
        else:
            raise RuntimeError('cell complex has not been constructed')

    @staticmethod
    def _obj_str(cells, use_mtl=False, filename_mtl='colours.mtl'):
        """
        Convert a list of cells into a string of obj format.

        Parameters
        ----------
        cells: list of Polyhedra objects
            Polyhedra cells
        use_mtl: bool
            Use mtl attribute in obj if set True
        filename_mtl: None or str
            Material filename

        Returns
        -------
        scene_str: str
            String representation of the object
        material_str: str
            String representation of the material
        """
        scene = None
        for cell in cells:
            scene += cell.render_solid()

        # directly save the obj string from scene.obj() will bring the inverted facets
        scene_obj = scene.obj_repr(scene.default_render_params())
        if len(cells) == 1:
            scene_obj = [scene_obj]
        scene_str = ''
        material_str = ''

        if use_mtl:
            scene_str += f'mtllib {filename_mtl}\n'

        for o in range(len(cells)):
            scene_str += scene_obj[o][0] + '\n'

            if use_mtl:
                scene_str += scene_obj[o][1] + '\n'
                material_str += 'newmtl ' + scene_obj[o][1].split()[1] + '\n'
                material_str += 'Kd {:.3f} {:.3f} {:.3f}\n'.format(random(), random(), random())  # diffuse colour

            scene_str += '\n'.join(scene_obj[o][2]) + '\n'
            scene_str += '\n'.join(scene_obj[o][3]) + '\n'  # contents[o][4] are the interior facets
        return scene_str, material_str

    def save_obj(self, filepath, indices_cells=None, use_mtl=False):
        """
        Save polygon soup of indexed convexes to an obj file.

        Parameters
        ----------
        filepath: str or Path
            Filepath to save obj file
        indices_cells: (n,) int
            Indices of cells to save to file
        use_mtl: bool
            Use mtl attribute in obj if set True
        """
        if self.constructed:
            # create the dir if not exists
            filepath = Path(filepath)
            filepath.parent.mkdir(parents=True, exist_ok=True)

            cells = [self.cells[i] for i in indices_cells] if indices_cells is not None else self.cells
            scene_str, material_str = self._obj_str(cells, use_mtl=use_mtl, filename_mtl=f'{filepath.stem}.mtl')

            with open(filepath, 'w') as f:
                f.writelines(scene_str)
            if use_mtl:
                with open(filepath.with_name(f'{filepath.stem}.mtl'), 'w') as f:
                    f.writelines(material_str)
        else:
            raise RuntimeError('cell complex has not been constructed')

    def save_plm(self, filepath, indices_cells=None):
        """
        Save polygon soup of indexed convexes to a plm file (polyhedron mesh in Mapple).

        Parameters
        ----------
        filepath: str or Path
            Filepath to save plm file
        indices_cells: (n,) int
            Indices of cells to save to file
        """
        if self.constructed:
            # create the dir if not exists
            filepath = Path(filepath)
            filepath.parent.mkdir(parents=True, exist_ok=True)

            num_vertices = 0
            info_vertices = ''
            info_facets = ''
            info_header = ''

            cells = [self.cells[i] for i in indices_cells] if indices_cells is not None else self.cells

            scene = None
            for cell in cells:
                scene += cell.render_solid()
                num_vertices += cell.n_vertices()

            info_header += '#vertices {}\n'.format(num_vertices)
            info_header += '#cells {}\n'.format(len(cells))

            with open(filepath, 'w') as f:
                contents = scene.obj_repr(scene.default_render_params())
                for o in range(len(cells)):
                    info_vertices += '\n'.join([st[2:] for st in contents[o][2]]) + '\n'
                    info_facets += str(len(contents[o][3])) + '\n'
                    for st in contents[o][3]:
                        info_facets += str(len(st[2:].split())) + ' '  # number of vertices on this facet
                        info_facets += ' '.join([str(int(n) - 1) for n in st[2:].split()]) + '\n'
                f.writelines(info_header + info_vertices + info_facets)

        else:
            raise RuntimeError('cell complex has not been constructed')



import numpy as np
from pathlib import Path

from abspy import VertexGroup, VertexGroupReference

dir_tests = Path(__file__).parent


def example_extract_primitives_vg():
    """
    Extract primitives from VertexGroup (.vg) file.
    """
    vertex_group = VertexGroup(filepath=dir_tests / 'test_data' / 'test_points.vg')

    # append additional planes
    additional_planes = [[1, 2, 3, 4], [5, 6, 7, 8]]
    additional_points = [[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                         [[10, 11, 12], [13, 14, 15], [16, 17, 18]]]
    vertex_group.append_planes(additional_planes=additional_planes, additional_points=additional_points)

    # save planes and bounds as numpy (.npy) files
    vertex_group.save_planes_npy(dir_tests / 'test_output' / 'primitives_planes_vg.npy')
    vertex_group.save_aabbs_npy(dir_tests / 'test_output' / 'primitives_bounds_vg.npy')

    # save extracted primitives to a Vertex group (.vg) file
    vertex_group.save_vg(dir_tests / 'test_output' / 'primitives.vg')


def example_extract_primitives_bvg():
    """
    Extract primitives from VertexGroup (.vg) file.
    """
    vertex_group = VertexGroup(filepath=dir_tests / 'test_data' / 'test_points.bvg')

    # append additional planes
    additional_planes = [[1, 2, 3, 4], [5, 6, 7, 8]]
    additional_points = [[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                         [[10, 11, 12], [13, 14, 15], [16, 17, 18]]]
    vertex_group.append_planes(additional_planes=additional_planes, additional_points=additional_points)

    # save planes and bounds as numpy (.npy) files
    vertex_group.save_planes_npy(dir_tests / 'test_output' / 'primitives_planes_bvg.npy')
    vertex_group.save_aabbs_npy(dir_tests / 'test_output' / 'primitives_bounds_bvg.npy')

    # save extracted primitives to a binary Vertex group (.bvg) file
    vertex_group.save_bvg(dir_tests / 'test_output' / 'primitives.bvg')


def example_extract_reference_primitives():
    """
    Extract primitives from VertexGroupReference (.ply) file.
    """
    vertex_group_reference = VertexGroupReference(filepath=dir_tests / 'test_data' / 'test_mesh.ply', num_samples=10000)

    # inject points
    vertex_group_reference.inject_points(np.random.rand(100, 3) - 0.5, threshold=0.05, overwrite=True,
                                         keep_bottom=True, keep_wall=True, compute_normal=True,
                                         pseudo_normal=False, pseudo_size=0)

    # save point cloud
    vertex_group_reference.save_cloud(dir_tests / 'test_output' / 'reference.ply')

    # save extracted primitives to both a Vertex Group (.vg) file and a binary Vertex group (.bvg) file
    vertex_group_reference.save_vg(dir_tests / 'test_output' / 'reference.vg')
    vertex_group_reference.save_bvg(dir_tests / 'test_output' / 'reference.bvg')


if __name__ == '__main__':
    example_extract_primitives_vg()
    example_extract_primitives_bvg()
    example_extract_reference_primitives()



from pathlib import Path

from abspy import VertexGroup, AdjacencyGraph, CellComplex
import numpy as np
import random

random.seed(100)

dir_tests = Path(__file__).parent


def sigmoid(x):
    # can safely ignore RuntimeWarning: overflow encountered in exp
    return 1 / (1 + np.exp(-x))


def example_combined():
    """
    Full workflow from VertexGroup to reconstructed surface.
    """
    # load a point cloud in VertexGroup
    vertex_group = VertexGroup(filepath=dir_tests / 'test_data' / 'test_points.vg')

    # normalise the point cloud
    vertex_group.normalise_to_centroid_and_scale()

    # additional planes to append (e.g., the bounding planes), this example does not apply
    additional_planes = [[0, 0, 1, -vertex_group.aabbs[:, 0, 2].min()]]  # the bottom of the points (z = d)

    # initialise CellComplex from planar primitives
    cell_complex = CellComplex(vertex_group.planes, vertex_group.aabbs, vertex_group.obbs, vertex_group.points_grouped,
                               build_graph=True, additional_planes=None)

    # refine planar primitives
    cell_complex.refine_planes(theta=0.1745, epsilon=0.005)

    # prioritise certain planes (e.g., vertical ones), this example does not apply
    cell_complex.prioritise_planes(prioritise_verticals=False)

    # construct CellComplex
    cell_complex.construct()

    # print info on the cell complex
    cell_complex.print_info()

    # cells inside reference mesh
    cells_in_mesh = cell_complex.cells_in_mesh(dir_tests / 'test_data' / 'test_mesh.ply')

    # visualise the inside cells (only if pyglet installation is found and valid indices are provided)
    if len(cells_in_mesh):
        try:
            cell_complex.visualise(indices_cells=cells_in_mesh, temp_dir=str(dir_tests) + '/test_output/')
        except (Exception, ImportError) as e:
            print(f'visualization skipped: {e}')

    # build adjacency graph of the cell complex
    adjacency_graph = AdjacencyGraph(cell_complex.graph)

    # assign weights to n-links and st-links to the graph
    adjacency_graph.assign_weights_to_n_links(cell_complex.cells, attribute='area_overlap',
                                              factor=0.001, cache_interfaces=True)
    adjacency_graph.assign_weights_to_st_links(
        {c: 0.8 if c in adjacency_graph.to_uids(cells_in_mesh) else 0.2 for c in adjacency_graph.uid})

    # perform graph-cut to extract surface
    _, _ = adjacency_graph.cut()

    # save surface model to an OBJ file
    adjacency_graph.save_surface_obj(filepath=dir_tests / 'test_output' / 'surface.obj', engine='mesh')


if __name__ == '__main__':
    example_combined()
